#ifndef _H5SHORTNAME_H
#define _H5SHORTNAME_H

#include "config_hdf5.h"

#include <map>
#include <sstream>
#include <string>
#include <vector>
#include <debug.h>

using namespace std;

/// A class for keeping track of short names.
///
/// This class contains functions that generate a shortened variable names
/// to help visualization clients like GrADS. The "sdfopen" command in
/// GrADS expects that all variables are less than 15 characters long.
/// Otherwise, you need to use "xdfopen" command in GrADS which requires
/// additional definition file that maps long name to short one.
/// 
///
/// @author Hyo-Kyung Lee <hyoklee@hdfgroup.org>
///
/// Copyright (c) 2009 The HDF Group
///
/// All rights reserved.
class H5ShortName {
  
public:
    /// This variable is incremented to generate unique id that will be
    // appended to "A" character.
    int index;
    /// Remembers the long name to short name mapping.
    map < string, string > long_to_short;
    
    /// Remembers the short name to long name mapping.
    map < string, string > short_to_long;	
  
    H5ShortName();

    /// Drops the group path name from \a a_name and returns the dataset name
    /// only by searching for the last instance of '/' character.
    /// 
    /// \param a_name  HDF-EOS5 dataset name
    /// \see H5CF.cc
    /// \see H5EOS.cc
    /// \return substring 
    string cut_long_name(string a_name);

    /// Generates a unique name from \a a_name
    ///
    /// This function generates  a new name that starts with "A" and a unique
    /// id number that starts from 0. It cuts the \a a_name to make the final
    /// string shorter than 15 characters long.
    /// 
    /// \param a_name  HDF-EOS5 dataset name
    /// \see H5CF.cc
    /// \see H5EOS.cc
    /// \return a new, 15-character long string.
    string generate_short_name(string a_name);

    /// Retrieves a long name from \a a_short_name
    ///
    /// \param a_short_name generated by CF option
    ///
    /// \return a string that is the original full-path dataset name.
    string get_long_name(string a_short_name);

    /// Retrieves a short name from \a a_long_name
    ///
    /// \param a_long_name a full-path dataset name.
    ///
    /// \return a shortened string generated by CF option
    string get_short_name(string a_long_name);

    ///  Clears all internal map variables and makes index 0.
    void reset();
};

#endif
